"""Utilities for cheating on Wordle."""
try:
    from importlib.metadata import version, PackageNotFoundError

except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError


try:
    __version__ = version(__name__)

except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



"""Command-line interface for wordle-cheater."""
import click
from wordle_cheater.cheater import cheat, get_wordle_letters
from wordle_cheater.interface import ClickInterface, CursesInterface
from wordle_cheater.interface_base import format_words


@click.command()
@click.argument("words", default="")
@click.argument("colors", default="")
@click.option(
    "--print/--no-print",
    "-p",
    "print_",
    default=False,
    show_default=True,
    help="When interactively entering guesses, print guesses and solutions on exit.",
)
@click.option(
    "--rows",
    default=10,
    show_default=True,
    help="Maximum rows of possible solutions to print.",
)
@click.option(
    "--cols",
    default=8,
    show_default=True,
    help="Maximum columns of possible solutions to print.",
)
@click.option(
    "--simple-print/--no-simple-print",
    default=False,
    show_default=True,
    help="Print all solutions, separated by a space.  Useful for machine-parsing.",
)
@click.option(
    "--use-curses/--no-curses",
    default=True,
    show_default=True,
    help="Use the Curses library for interactive input and output.",
)
def wordle_cheat(words, colors, print_, rows, cols, simple_print, use_curses):
    """Cheat on Wordle :(

    Given your current guesses (WORDS) and their colors (COLORS), this utility prints
    a list of possible solutions in random order.  If either WORDS or COLORS are not
    provided, wordle-cheater has you interactively enter your guesses.

    COLORS must be a string of 'b', 'y', or 'g' characters, corresponding to black,
    yellow, and green, respectively.

    WORDS and COLORS both ignore whitespace.

    When interactively entering guesses, press space once before a letter to mark it as
    yellow, or press space twice to mark it as green.

    Note that no-curses mode is poorly supported and may not work on your terminal.

    """  # noqa: D400 DAR101
    # If words & colors were already given, no need to spawn a UI
    if words != "" and colors != "":
        guesses = get_wordle_letters(words, colors)
        print_ = True  # Ensure we print results

    elif use_curses:  # pragma: no cover
        ui = CursesInterface.init_and_run()
        guesses = ui.guesses

    else:  # pragma: no cover
        ui = ClickInterface(max_rows=rows, max_cols=cols)
        ui.main()
        return  # ClickInterface prints solutions automatically, so we're done

    # Get possible words
    possible_words = cheat(guesses)

    # Now print the entered guesses and results, if requested
    if simple_print:
        out_str = " ".join(possible_words)
        click.echo(out_str)

    elif print_:
        out_str = format_words(possible_words, max_rows=rows, max_cols=cols)
        click.secho("Wordle Cheater :(", bold=True)
        click.echo()  # Get a new line
        click.echo("    ", nl=False)  # Indent by four spaces
        for wl in guesses:
            if wl.color == "black":
                # Have to treat black differently since the fg needs to be white
                click.secho(wl.letter.upper(), fg="white", bg="black", nl=False)

            else:
                click.secho(wl.letter.upper(), fg="black", bg=wl.color, nl=False)

            # If we're at the end of the word (5 letters), go to the next line
            if wl.index == 4:
                click.echo()  # Get a new line
                click.echo("    ", nl=False)  # Indent by four spaces

        click.echo()
        click.secho("Possible solutions:", underline=True)
        click.secho(out_str)

    else:  # pragma: no cover
        return


if __name__ == "__main__":
    wordle_cheat()  # pragma: no cover



"""Utilities for generating solutions to Wordle given previous guesses."""
from dataclasses import dataclass

from wordle_cheater.dictionary import letters
from wordle_cheater.dictionary import wordle_words


@dataclass
class WordleLetter:
    """Class describing a single letter in a Wordle guess.

    Parameters
    ----------
    letter : str
        The letter in question.  Should be lower case.
    color : {"black", "yellow", "green"}
        The color Wordle marked the letter.
    index : {0, 1, 2, 3, 4}
        The location in the word where the letter appeared.
    """

    letter: str
    color: str
    index: int

    def __post_init__(self):
        """Do parameter validation."""
        if self.letter.upper() not in letters:
            raise ValueError("`letter` must be an english letter.")

        if not self.letter.lower() == self.letter:
            raise ValueError("`letter` must be lowercase.")

        if self.color not in ["black", "yellow", "green"]:
            raise ValueError("`color` must be one of ['black', 'yellow', 'green']")

        if not (isinstance(self.index, int)):
            raise ValueError("`index` must be integer in range [0, 5)")

        if self.index not in range(5):
            raise ValueError("`index` must be integer in range [0, 5)")

    def __lt__(self, other):
        """Deterimine if `self` is less than `other`.

        Sorts by index.

        Parameters
        ----------
        other : WordleLetter
            The other WordleLetter to compare against.

        Returns
        -------
        bool
            Whether `self.index` is less than `other.index`.
        """
        return self.index < other.index


class InvalidWordleLetters(Exception):
    """Exception for when invalid letters are passed to parse_worlde_letters.

    Attributes
    ----------
    invalid_letters : list of WordleLetter objects
        The relevant letters that were found to be invalid.
    """

    def __init__(self, message, wordle_letters):
        self.invalid_letters = wordle_letters
        super().__init__(message)


def _flatten(list_):
    """Flatten a list of lists into a 1D list."""
    return [item for sublist in list_ for item in sublist]


def check_word(
    word,
    blacks=None,
    yellows=None,
    greens=None,
    counts=None,
    check_dict=True,
):
    """Check if `word` is a possible solution given previous guesses.

    All inputs must be lowercase as we don't bother to cast them to lowercase to save a
    bit of time.

    Parameters
    ----------
    word : length-5 string
        The word to check.
    blacks : length-5 list of lists, optional
        A list of lowercase letters that are not in the word.  For example, if our
        guesses have the letter 'A' marked black at the second character,
        `blacks = [[], ['A'], [], [], []]`.  Defaults to no letters marked black.
    yellows : length-5 list of lists, optional
        Lowercase letters that are in the word, but not in the correct location.  For
        example, if our guesses tell us that the letter 'A' was in the word, but it was
        not the third letter, we
        would pass `yellows = [[], [], ['a'], [], []]`.  Defaults to no letters marked
        yellow.
    greens : length-5 list, optional
        Lowercase letters that are in the word and in the correct location.  For
        example, if our guesses tell us that the letter 'A' is the fourth letter of the
        word, we would pass `greens = [None, None, None, 'a', None]`.  Defaults to no
        letters marked green.
    counts : dict, optional
        Counts of letters that should appear in the solution.  For letters that are in
        `blacks`, this is interpreted as the exact number of times the letter must
        appear in the solution, and defaults to zero.  For letters that are in
        `yellows` and/or `greens`, this is interpreted as the minimum number of times
        the letter must appear in the solution, and defaults to one.
        For example, if a previous guess was 'array' with the two 'r's colored, then we
        would know the solution must have at least two 'r's and pass
        `counts = {'r': 2}`.
        If a previous guess was 'array', with one 'r' marked black and one colored, then
        we know the solution must have exactly one 'r' and pass `counts = {'r': 1}`.
    check_dict : bool, optional
        Whether or not to check if `word` is a real five letter english word.

    Returns
    -------
    valid : bool
        Whether or not `word` is a possible solution given `blacks`, `yellows`, and
        `greens`.
    """
    if blacks is None:
        blacks = [[], [], [], [], []]

    if yellows is None:
        yellows = [[], [], [], [], []]

    if greens is None:
        greens = [None, None, None, None, None]

    if counts is None:
        counts = dict()

    assert len(blacks) == 5
    assert len(yellows) == 5
    assert len(greens) == 5

    # Get unraveled lists
    all_blacks = _flatten(blacks)
    all_yellows = _flatten(yellows)

    # Check for "hard mode compliance"
    known_letters = all_yellows + [l_ for l_ in greens if l_ is not None]
    for known_letter in known_letters:
        if known_letter not in word:
            return False

        # Also check that there are enough repetitions if applicable
        elif word.count(known_letter) < counts.get(known_letter, 1):
            return False

    # Now check each letter for compatibility with known information
    for i, letter in enumerate(word):
        if letter in all_blacks:
            # If a letter appears as black in the previous guesses, we know how many
            # of that letter we must have - check that here.
            n_of_letter_max = counts.get(letter, 0)
            n_of_letter_in_word = word.count(letter)

            if n_of_letter_in_word > n_of_letter_max:
                return False

        if letter in blacks[i]:
            return False

        elif letter in yellows[i]:
            return False

        elif letter != greens[i] and greens[i] is not None:
            return False

    # Check if guess is a real word
    if check_dict:
        if word not in wordle_words:
            return False

    # If we've made it this far, the word is a possible solution
    return True


def find_words(blacks=None, yellows=None, greens=None, counts=None):
    """Find all possible words that are consistent with current information.

    Parameters
    ----------
    blacks : length-5 list of lists, optional
        A list of lowercase letters that are not in the word.  For example, if our
        guesses have the letter 'A' marked black at the second character,
        `blacks = [[], ['A'], [], [], []]`.  Defaults to no letters marked black.
    yellows : length-5 list of lists, optional
        Lowercase letters that are in the word, but not in the correct location.  For
        example, if our guesses tell us that the letter 'A' was in the word, but it was
        not the third letter, we
        would pass `yellows = [[], [], ['a'], [], []]`.  Defaults to no letters marked
        yellow.
    greens : length-5 list, optional
        Lowercase letters that are in the word and in the correct location.  For
        example, if our guesses tell us that the letter 'A' is the fourth letter of the
        word, we would pass `greens = [None, None, None, 'a', None]`.  Defaults to no
        letters marked green.
    counts : dict, optional
        Counts of letters that should appear in the solution.  For letters that are in
        `blacks`, this is interpreted as the exact number of times the letter must
        appear in the solution, and defaults to zero.  For letters that are in
        `yellows` and/or `greens`, this is interpreted as the minimum number of times
        the letter must appear in the solution, and defaults to one.
        For example, if a previous guess was 'array' with the two 'r's colored, then we
        would know the solution must have at least two 'r's and pass
        `counts = {'r': 2}`.  If a previous guess was 'array', with one 'r' marked black
        and one colored, then
        we know the solution must have exactly one 'r' and pass `counts = {'r': 1}`.

    Returns
    -------
    possible_words : list
        List of possible solutions that match the given guesses.
    """
    possible_words = []
    for word in wordle_words:
        if check_word(
            word,
            blacks=blacks,
            yellows=yellows,
            greens=greens,
            counts=counts,
        ):
            possible_words.append(word)

    return possible_words


def parse_wordle_letters(wordle_letters):
    """Parse and validate a list of WordleLetter objects.

    Parameters
    ----------
    wordle_letters : list of WordleLetter objects
        The current guesses, a list of WordleLetter objects.  Must be from an integer
        number of words, so `len(wordle_letters)` must be an integer multiple of 5.

    Returns
    -------
    blacks : length-5 list of lists
        A list of lowercase letters that are not in the word.  For example, if our
        guesses have the letter 'A' marked black at the second character,
        `blacks = [[], ['A'], [], [], []]`.
    yellows : length-5 list of lists
        Lowercase letters that are in the word, but not in the correct location.  For
        example, if our guesses tell us that the letter 'A' was in the word, but it was
        not the third letter, `yellows = [[], [], ['a'], [], []]`.
    greens : length-5 list
        Lowercase letters that are in the word and in the correct location.  For
        example, if our guesses tell us that the letter 'A' is the fourth letter of the
        word, `greens = [None, None, None, 'a', None]`.
    counts : dict
        Counts of letters that should appear in the solution.  For letters that are in
        `blacks`, this is interpreted as the exact number of times the letter must
        appear in the solution, and defaults to zero.  For letters that are in
        `yellows` and/or `greens`, this is interpreted as the minimum number of times
        the letter must appear in the solution, and defaults to one.
        For example, if a previous guess was 'array' with the two 'r's colored, then we
        would know the solution must have at least two 'r's and so `counts = {'r': 2}`.
        If a previous guess was 'array', with one 'r' marked black and one colored, then
        we know the solution must have exactly one 'r' and so `counts = {'r': 1}`.

    # noqa : DAR000
    """
    blacks = [[], [], [], [], []]
    yellows = [[], [], [], [], []]
    greens = [None, None, None, None, None]
    counts = dict()

    if len(wordle_letters) % 5 != 0:
        raise ValueError("`len(wordle_letters)` must be an integer multiple of five")

    words = [wordle_letters[i * 5 : i * 5 + 5] for i in range(len(wordle_letters) // 5)]

    for word in words:
        these_blacks = [wl for wl in word if wl.color == "black"]
        these_yellows = [wl for wl in word if wl.color == "yellow"]
        these_greens = [wl for wl in word if wl.color == "green"]

        these_counts = dict()
        for wl in these_yellows + these_greens:
            these_counts[wl.letter] = these_counts.get(wl.letter, 0) + 1

        invalid_letters = []  # For letters incompatible with previous words

        # Validate black letters
        for wl in these_blacks:
            # A black letter cannot be colored in this word fewer times than in any
            # previous word, and can't have been previously marked colored at this
            # location
            curr_count = these_counts.get(wl.letter, 0)
            prev_count = counts.get(wl.letter, 0)
            if (
                curr_count < prev_count
                or wl.letter in yellows[wl.index]
                or wl.letter == greens[wl.index]
            ):
                invalid_letters.append(wl)

        # Validate yellow letters
        for wl in these_yellows:
            # A yellow letter can't have been previously marked black or green in
            # this location, and if it was previously marked black (anywhere), it must
            # have also been previously colored
            if (
                wl.letter in blacks[wl.index]
                or wl.letter == greens[wl.index]
                or wl.letter in _flatten(blacks)
                and counts.get(wl.letter, 0) == 0
            ):
                invalid_letters.append(wl)

        # Validate green letters
        for wl in these_greens:
            # A green letter can't have been previously marked black or yellow in
            # this location, a different letter can't have been marked green in this
            # location, and if it was previously marked black (anywhere), it must have
            # also been previously colored
            if (
                wl.letter in blacks[wl.index]
                or wl.letter in yellows[wl.index]
                or wl.letter != greens[wl.index]
                and greens[wl.index] is not None
                or wl.letter in _flatten(blacks)
                and counts.get(wl.letter, 0) == 0
            ):
                invalid_letters.append(wl)

        # Raise an error, if necessary
        if len(invalid_letters) > 0:
            invalid_letters = sorted(invalid_letters)  # Sort by index
            letters_str = ", ".join([wl.letter.upper() for wl in invalid_letters])
            inds_str = ", ".join([str(wl.index) for wl in invalid_letters])
            exc_str = (
                "Letters "
                + letters_str
                + " (indices "
                + inds_str
                + ") incompatible with previous entries"
            )
            raise InvalidWordleLetters(exc_str, invalid_letters)

        # If we made it through all that validation, append these letters to output
        for wl in word:
            if wl.color == "black":
                blacks[wl.index].append(wl.letter)

            elif wl.color == "yellow":
                yellows[wl.index].append(wl.letter)

            else:  # Color is green
                greens[wl.index] = wl.letter

            # Also update counts if necessary
            # If not playing hard mode, the current count of a letter could be less
            # than the final count, so only update count if it has increased
            if these_counts.get(wl.letter, 0) > counts.get(wl.letter, 0):
                counts[wl.letter] = these_counts[wl.letter]

    return blacks, yellows, greens, counts


def cheat(wordle_letters):
    """Given a list of WordleLetter objects, return possible solutions.

    Parameters
    ----------
    wordle_letters : list of WordleLetter
        The previous guesses.  Every five WordleLetters are interpreted as one word.

    Returns
    -------
    possible_words : list of str
        A list of the possible solutions given `wordle_letters`, in random order.
    """
    blacks, yellows, greens, counts = parse_wordle_letters(wordle_letters)
    possible_words = find_words(blacks, yellows, greens, counts)
    return possible_words


def get_wordle_letters(letters, colors):
    """Create a list of WordleLetter objects.

    Parameters
    ----------
    letters : str
        The previous guesses.  Whitespace will be ignored.
    colors : str
        The colors corresponding to each letter.  Each character in this string must
        be one of 'b', 'y', or 'g', which correspond to 'black', 'yellow', or 'green',
        respectively.  Whitespace will be ignored.

    Returns
    -------
    wordle_letters : list of WordleLetter
        The WordleLetter objects corresponding to the input letters and colors.
    """
    # Remove whitespace
    letters = "".join(letters.split())
    colors = "".join(colors.split())

    # Make lower case
    letters = letters.lower()
    colors = colors.lower()

    # Create WordleLetter objects
    wordle_letters = []
    for i, (l, c) in enumerate(zip(letters, colors)):
        if c == "b":
            color = "black"

        elif c == "y":
            color = "yellow"

        elif c == "g":
            color = "green"

        else:
            raise ValueError(f"`colors` must contain only 'b', 'y', or 'g' (got {c}).")

        wordle_letters.append(WordleLetter(letter=l, color=color, index=i % 5))

    return wordle_letters


def easy_cheat(letters, colors):
    """Given previous guesses on Wordle, return a list of possible solutions.

    Parameters
    ----------
    letters : str
        The previous guesses.  Whitespace will be ignored.
    colors : str
        The colors corresponding to each letter.  Each character in this string must
        be one of 'b', 'y', or 'g', which correspond to 'black', 'yellow', or 'green',
        respectively.  Whitespace will be ignored.

    Returns
    -------
    possible_words : list of str
        A list of the possible solutions in random order.
    """
    wordle_letters = get_wordle_letters(letters, colors)
    possible_words = cheat(wordle_letters)
    return possible_words



"""Interfaces for interactively entering guesses."""
import curses
import time

import click
from wordle_cheater.interface_base import WordleCheaterUI


class CursesInterface(WordleCheaterUI):
    """Interface for using the curses library to enter guesses and display solutions.

    Attributes
    ----------
    guesses : list of WordleLetter objects
        The currently entered guesses.
    entering_letters : bool
        Whether or not we are currently entering guesses.
    """

    @classmethod
    def init_and_run(cls, *args, **kwargs):
        """Instantiate and run `self.main()` using `curses.wrapper`.

        Parameters
        ----------
        *args : tuple
            Positional arguments to be passed to the CursesInterface constructor.
        **kwargs : dict, optional
            Keyword arguments to be passed to the CursesInterface constructor.

        Returns
        -------
        CursesInterface object
            An instance of the CursesInterface class.
        """
        ui = cls(*args, **kwargs)
        curses.wrapper(ui.main)
        return ui

    def main(self, stdscr):
        """Run the interface.

        Should typically be called using `curses.wrapper`.

        Parameters
        ----------
        stdscr : curses.Window object
            The curses screen which the user interacts with.
        """
        self.stdscr = stdscr
        curses.use_default_colors()
        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_BLACK)  # White on black
        curses.init_pair(2, curses.COLOR_BLACK, curses.COLOR_YELLOW)  # Black on yellow
        curses.init_pair(3, curses.COLOR_BLACK, curses.COLOR_GREEN)  # Black on green
        curses.init_pair(4, curses.COLOR_BLACK, curses.COLOR_RED)  # Black on red

        height, width = stdscr.getmaxyx()
        self.results_window = curses.newwin(
            height - 12, width, 12, 0
        )  # window for printing results

        x0 = width // 2 - 3
        y0 = 5
        self.print_title()
        self.enter_letters(x0=x0, y0=y0)
        self.print_results()
        self.set_cursor_visibility(False)
        self.get_key()

    def center_print(self, y, string, *args, **kwargs):
        """Print in the center of the screen.

        Parameters
        ----------
        y : int
            The vertical location at which to print.
        string : str
            The string to print.
        *args : tuple
            Additional arguments to be passed to `stdscr.addstr`.
        **kwargs : dict, optional
            Keyword arguments to be passed to `stdscr.addstr`.
        """
        height, width = self.stdscr.getmaxyx()

        str_length = len(string)
        x_mid = width // 2

        self.stdscr.addstr(y, x_mid - str_length // 2, string, *args, **kwargs)

    def print_title(self):
        """Print title and instructions."""
        self.center_print(1, "Wordle Cheater :(", curses.A_BOLD)
        self.center_print(2, "Enter guesses below.")
        self.center_print(3, "spacebar: change color", curses.A_DIM)

    def print_results(self, sep="     "):
        """Print possible solutions given guesses."""
        height, width = self.results_window.getmaxyx()
        max_rows = height - 1  # -1 to account for "Possible solutions" header
        cols = width // (5 + len(sep))

        out_str = self.get_results_string(max_rows=max_rows, max_cols=cols, sep=sep)

        self.results_window.clear()
        self.results_window.addstr(0, 0, "Possible solutions:", curses.A_UNDERLINE)
        self.results_window.addstr(1, 0, out_str)
        self.results_window.refresh()

    def print(self, x, y, string, c=None):
        """Print a string at coordinates x, y.

        Parameters
        ----------
        x : int
            Horizontal position at which to print the string.
        y : int
            Height at which to print the string.
        string : str
            The string to print.
        c : str, {None, 'black', 'yellow', 'green', 'red'}
            The color in which to print.  Must be one of
            ['black', 'yellow', 'green', 'red'] or None. If `c` is None, it should
            print in the default color pair.
        """
        if c is None:
            self.stdscr.addstr(y, x, string)

        elif c == "black":
            self.stdscr.addstr(y, x, string, curses.color_pair(1))

        elif c == "yellow":
            self.stdscr.addstr(y, x, string, curses.color_pair(2))

        elif c == "green":
            self.stdscr.addstr(y, x, string, curses.color_pair(3))

        elif c == "red":
            self.stdscr.addstr(y, x, string, curses.color_pair(4))

        else:
            raise ValueError(
                "`c` must be one of ['black', 'yellow', 'green', 'red'] or none."
            )

    def sleep(self, ms):
        """Temporarily suspend execution.

        Parameters
        ----------
        ms : int
            Number of miliseconds before execution resumes.
        """
        curses.napms(ms)
        self.stdscr.refresh()

    def move_cursor(self, x, y):
        """Move cursor to position x, y.

        Parameters
        ----------
        x : int
            Desired horizontal position of cursor.
        y : int
            Desired vertical position of cursor.
        """
        self.stdscr.move(y, x)

    def set_cursor_visibility(self, visible):
        """Set cursor visibility.

        Parameters
        ----------
        visible : bool
            Whether or not the cursor is visible.
        """
        curses.curs_set(visible)

    def get_key(self):
        """Get a key press.

        Returns
        -------
        key : str
            The key that was pressed.
        """
        return self.stdscr.getkey()

    def is_enter(self, key):
        """Check if `key` is the enter/return key.

        Parameters
        ----------
        key : str
            The key to check.

        Returns
        -------
        is_enter : bool
            True if `key` is the enter or return key, False otherwise.
        """
        if key == curses.KEY_ENTER or key == "\n" or key == "\r":
            return True

        else:
            return False

    def is_backspace(self, key):
        """Check if `key` is the backspace/delete key.

        Parameters
        ----------
        key : str
            The key to check.

        Returns
        -------
        is_backspace : bool
            True if `key` is the backspace or delete key, False otherwise.
        """
        if key == curses.KEY_BACKSPACE or key == "\b" or key == "\x7f":
            return True

        else:
            return False


class ClickInterface(WordleCheaterUI):
    """Interface for using Click alone to enter letters and see solutions.

    Parameters
    ----------
    max_rows : int, optional
        The maximum rows of possible solutions to print.
    max_cols : int, optional
        The maximum columns of possible solutions to print.
    x0 : int, optional
        The leftmost position where guesses will be entered.
    y0 : int, optional
        The topmost position where guesses will be entered.
    esc : str, optional
        The ANSI escape code for the terminal.

    Attributes
    ----------
    guesses : list of WordleLetter objects
        The currently entered guesses.
    entering_letters : bool
        Whether or not we are currently entering guesses.
    max_rows : int
        The maximum rows of possible solutions to print.
    max_cols : int
        The maximum columns of possible solutions to print.
    x0 : int
        The leftmost position where guesses will be entered.
    y0 : int
        The topmost position where guesses will be entered.
    esc : str
        The ANSI escape code for the terminal.
    line_lengths : list of int
        The highest x value we've printed to per line.  For example, if we've printed
        two lines, the first one up to x=5 and the second up to x=3, then
        `line_lengths = [5, 3]`.
    curs_xy
    """

    def __init__(self, max_rows=10, max_cols=8, x0=4, y0=4, esc="\033"):
        self.max_rows = max_rows  # Maximum rows of results to print
        self.max_cols = max_cols  # Maximum columns of results to print
        self.x0 = x0  # Initial x position of guesses
        self.y0 = y0  # Initial y position of guesses
        self.esc = esc  # ANSI escape code
        self._curs_xy = (0, 0)  # cursor position
        self.line_lengths = [0]  # Highest x values we've hit per line
        super().__init__()

    @property
    def curs_xy(self):
        """Location of cursor."""
        return self._curs_xy

    @curs_xy.setter
    def curs_xy(self, xy):
        """Update max line lengths when we update cursor position."""
        x, y = xy
        if y > len(self.line_lengths) - 1:
            self.line_lengths += [0 for i in range(y - len(self.line_lengths) + 1)]

        if x > self.line_lengths[y]:
            self.line_lengths[y] = x

        self._curs_xy = xy

    def main(self):
        """Run the interface."""
        try:
            self.print_title()
            self.enter_letters(x0=self.x0, y0=self.y0)
            self.print_results()

        finally:
            self.set_cursor_visibility(True)

    def print_title(self):
        """Print title and instructions."""
        self.print(0, 0, "Wordle Cheater :(", bold=True)
        self.print(0, 1, "Enter guesses below.")
        self.print(0, 2, "spacebar: change color", dim=True)

    def print_results(self):
        """Print possible solutions given guesses."""
        # If we're still entering letters, don't do anything
        if self.entering_letters:
            return

        out_str = self.get_results_string(
            max_rows=self.max_rows, max_cols=self.max_cols, sep="     "
        )

        self.move_cursor(0, self.curs_xy[1] + 1)
        click.secho("Possible solutions:", underline=True)
        click.echo(out_str)

    def print(self, x, y, string, c=None, *args, **kwargs):
        """Print a string at coordinates x, y.

        Parameters
        ----------
        x : int
            Horizontal position at which to print the string.
        y : int
            Height at which to print the string.
        string : str
            The string to print.
        c : str, {None, 'black', 'yellow', 'green', 'red'}
            The color in which to print.  Must be one of
            ['black', 'yellow', 'green', 'red'] or None. If `c` is None, it should
            print in the default color pair.
        *args : tuple
            Additional arguments to be passed to `click.secho`.
        **kwargs : dict, optional
            Keyword arguments to be passed to `click.secho`.
        """
        # Move cursor to x, y so we can print there
        self.move_cursor(x, y)

        if c is None:
            click.secho(string, nl=False, *args, **kwargs)

        elif c == "black":
            click.secho(string, fg="white", bg="black", nl=False)

        elif c == "yellow":
            click.secho(string, fg="black", bg="yellow", nl=False)

        elif c == "green":
            click.secho(string, fg="black", bg="green", nl=False)

        elif c == "red":
            click.secho(string, fg="black", bg="red", nl=False)

        else:
            raise ValueError(
                "`c` must be one of ['black', 'yellow', 'green', 'red'] or none."
            )

        self.curs_xy = (self.curs_xy[0] + len(string), self.curs_xy[1])

    def sleep(self, ms):
        """Temporarily suspend execution.

        Parameters
        ----------
        ms : int
            Number of miliseconds before execution resumes.
        """
        time.sleep(ms / 1000)

    def move_cursor(self, x, y):
        """Move cursor to position x, y.

        Parameters
        ----------
        x : int
            Desired horizontal position of cursor.
        y : int
            Desired vertical position of cursor.
        """
        # Check if we want to move cursor up (decreasing y)
        if self.curs_xy[1] > y:
            click.echo(f"{self.esc}[{self.curs_xy[1] - y}A", nl=False)

        # Check if we want to move cursor down (increasing y)
        elif self.curs_xy[1] < y:
            # Check if we need to add new lines to screen
            if len(self.line_lengths) - 1 < y:
                # First arrow down as far as possible
                click.echo(
                    f"{self.esc}[{(len(self.line_lengths) - 1) - self.curs_xy[1]}B",
                    nl=False,
                )

                # Now add blank lines
                click.echo("\n" * (y - (len(self.line_lengths) - 1)), nl=False)

                # New line, so definitely need to print spaces to move x
                click.echo(" " * x, nl=False)
                self.curs_xy = (x, y)
                return

            else:
                # Should just arrow down to not overwrite stuff
                click.echo(f"{self.esc}[{y - self.curs_xy[1]}B", nl=False)

        # Check if we want to move cursor left (decreasing x)
        if self.curs_xy[0] > x:
            click.echo(f"{self.esc}[{self.curs_xy[0] - x}D", nl=False)

        # Check if we want to move cursor right (increasing x)
        elif self.curs_xy[0] < x:
            # Check if we need to add space to right of cursor
            if self.line_lengths[y] > x:
                # First arrow to the right as far as possible
                click.echo(
                    f"{self.esc}[{self.line_lengths[y] - self.curs_xy[0]}C", nl=False
                )

                # Now add blank spaces
                click.echo(" " * (x - self.line_lengths[y]), nl=False)

            else:
                # Should just arrow to right to not overwrite stuff
                click.echo(f"{self.esc}[{x - self.curs_xy[0]}C", nl=False)

        self.curs_xy = (x, y)

    def set_cursor_visibility(self, visible):
        """Set cursor visibility.

        Parameters
        ----------
        visible : bool
            Whether or not the cursor is visible.
        """
        if visible:
            click.echo(f"{self.esc}[?25h", nl=False)

        else:
            click.echo(f"{self.esc}[?25l", nl=False)

    def get_key(self):
        """Get a key press.

        Returns
        -------
        key : str
            The key that was pressed.
        """
        return click.getchar()

    def is_enter(self, key):
        """Check if `key` is the enter/return key.

        Parameters
        ----------
        key : str
            The key to check.

        Returns
        -------
        is_enter : bool
            True if `key` is the enter or return key, False otherwise.
        """
        if key == "\r" or key == "\n":
            return True

        else:
            return False

    def is_backspace(self, key):
        """Check if `key` is the backspace/delete key.

        Parameters
        ----------
        key : str
            The key to check.

        Returns
        -------
        is_backspace : bool
            True if `key` is the backspace or delete key, False otherwise.
        """
        if key == "\b" or key == "\x7f":
            return True

        else:
            return False


if __name__ == "__main__":
    # curses_ui = CursesInterface()
    # curses.wrapper(curses_ui.main)
    click_ui = ClickInterface()
    click_ui.main()
